"""Django REST Framework like viewset mixins for common model sqlalchemy
actions."""

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.views.generic.edit import FormMixin

from ..db import meta
from ..forms import modelform_factory
from ..views.base import BaseMultipleObjectMixin, BaseSingleObjectMixin


class ListModelMixin(BaseMultipleObjectMixin):
    """A mixin for displaying a list of objects.

    When used with router, it will map the following operations to actions on the viewset

    ====== ======================== =============== ======================
    Method Path                     Action          Route Name
    ====== ======================== =============== ======================
    GET    /                        list            <resource name>-list
    ====== ======================== =============== ======================
    """

    def list(self, request, *args, **kwargs):
        """List action for displaying a list of objects."""
        self.object_list = self.get_queryset()
        context = self.get_list_context_data()
        return self.render_to_response(context)

    def get_list_context_object_name(self, object_list):
        """Get the name to use for the object."""
        model = self.get_model()
        return "%s_list" % model.__name__.lower()

    def get_list_context_data(self, **kwargs):
        """Returns context data for list action."""
        queryset = self.object_list
        page_size = self.get_paginate_by(queryset)
        context_object_name = self.get_list_context_object_name(queryset)
        if page_size:
            paginator, page, queryset, is_paginated = self.paginate_queryset(queryset, page_size)
            context = {"paginator": paginator, "page_obj": page, "is_paginated": is_paginated, "object_list": queryset}
        else:
            context = {"paginator": None, "page_obj": None, "is_paginated": False, "object_list": queryset}
        if context_object_name is not None:
            context[context_object_name] = queryset
        context.update(kwargs)
        return super().get_context_data(**context)


class RetrieveModelMixin(BaseSingleObjectMixin):
    """A mixin for displaying a single object.

    When used with router, it will map the following operations to actions on the viewset

    ====== ======================== =============== ======================
    Method Path                     Action          Route Name
    ====== ======================== =============== ======================
    GET    /<pk>/                   retrieve        <resource name>-detail
    ====== ======================== =============== ======================
    """

    def retrieve(self, request, *args, **kwargs):
        """List action for displaying a single object."""
        self.object = self.get_object()
        context = self.get_detail_context_data(object=self.object)
        return self.render_to_response(context)

    def get_url_kwargs(self, obj):
        info = meta.model_info(type(obj))
        return {key: getattr(obj, key) for key in info.primary_keys}

    def get_detail_context_object_name(self, obj):
        """Get the name to use for the object."""
        model = self.get_model()
        return model.__name__.lower()

    def get_detail_context_data(self, **kwargs):
        """Returns detail context data for template rendering."""
        context = {}
        if self.object:
            context["object"] = self.object
            context_object_name = self.get_detail_context_object_name(self.object)
            if context_object_name:
                context[context_object_name] = self.object
        context.update(kwargs)
        return super().get_context_data(**context)


class ModelFormMixin(FormMixin, RetrieveModelMixin):
    """Common mixin for handling sqlalchemy model forms in viewsets."""

    fields = None
    form_class = None
    success_url = None

    def get_form_class(self):
        """Returns the form class to be used."""
        if self.fields is not None and self.form_class:
            raise ImproperlyConfigured("Specifying both 'fields' and 'form_class' is not permitted.")

        if self.form_class:
            return self.form_class

        model = self.get_model()
        return modelform_factory(model, fields=self.fields, session=self.get_session())

    def get_form_kwargs(self):
        """Returns the keyword arguments for instantiating the form."""
        kwargs = super().get_form_kwargs()
        if hasattr(self, "object"):
            kwargs.update({"instance": self.object})

        kwargs["session"] = self.get_session()
        return kwargs

    def get_success_url(self):
        """Return the URL to redirect to after processing a valid form."""
        if self.success_url:
            return self.success_url.format(**vars(self.object))

        raise ImproperlyConfigured(
            "No URL to redirect to. Either provide a url or override this function to return a url"
        )

    def process_form(self, form):
        """Checks if form is valid and processes accordingly."""
        if form.is_valid():
            return self.form_valid(form)

        return self.form_invalid(form)

    def form_valid(self, form):
        """Processes a valid form."""
        self.object = form.save()
        return super().form_valid(form)

    def form_invalid(self, form):
        """Handles invalid form."""
        return self.render_to_response(self.get_form_context_data(form=form))

    def get_form_context_data(self, **kwargs):
        """Return form context data for template rendering."""
        if "form" not in kwargs:
            kwargs["form"] = self.get_form()
        return self.get_detail_context_data(**kwargs)


class CreateModelMixin(ModelFormMixin):
    """A mixin for supporting creating objects.

    When used with router, it will map the following operations to actions on the viewset

    ====== ======================== =============== ======================
    Method Path                     Action          Route Name
    ====== ======================== =============== ======================
    POST   /                        create          <resource name>-list
    GET    /new/                    new             <resource name>-new
    ====== ======================== =============== ======================
    """

    def new(self, request, *args, **kwargs):
        """New action for displaying a form for creating an object."""
        return self.render_to_response(self.get_create_context_data())

    def create(self, request, *args, **kwargs):
        """Create action for creating an object."""
        form = self.get_form()
        return self.process_form(form)

    def get_create_context_data(self, **kwargs):
        """Returns new context data for template rendering."""
        return self.get_form_context_data(**kwargs)


class UpdateModelMixin(ModelFormMixin):
    """A mixin for supporting updating objects.

    When used with router, it will map the following operations to actions on the viewset

    ====== ======================== =============== ======================
    Method Path                     Action          Route Name
    ====== ======================== =============== ======================
    GET    /<pk>/edit/              edit            <resource name>-edit
    POST   /<pk>/                   update          <resource name>-detail
    PUT    /<pk>/                   update          <resource name>-detail
    PATCH  /<pk>/                   update          <resource name>-detail
    ====== ======================== =============== ======================
    """

    def edit(self, request, *args, **kwargs):
        """Edit action for displaying a form for editing an object."""
        self.object = self.get_object()
        return self.render_to_response(self.get_update_context_data())

    def update(self, request, *args, **kwargs):
        """Update action for updating an object."""
        self.object = self.get_object()
        form = self.get_form()
        return self.process_form(form)

    def get_update_context_data(self, **kwargs):
        """Returns edit context data for template rendering."""
        return self.get_form_context_data(**kwargs)


class DeleteModelMixin(RetrieveModelMixin):
    """A mixin for supporting deleting objects.

    When used with router, it will map the following operations to actions on the viewset

    ====== ======================== =============== ======================
    Method Path                     Action          Route Name
    ====== ======================== =============== ======================
    GET    /<pk>/delete/            confirm_destoy  <resource name>-delete
    POST   /<pk>/delete/            destroy         <resource name>-delete
    DELETE /<pk>/                   destroy         <resource name>-detail
    ====== ======================== =============== ======================
    """

    destroy_success_url = None

    def confirm_destroy(self, request, *args, **kwargs):
        """Confirm_destory action for displaying deletion confirmation for an
        object."""
        self.object = self.get_object()
        return self.render_to_response(self.get_destroy_context_data())

    def destroy(self, request, *args, **kwargs):
        """Destroy action for deletion of an object."""
        self.object = self.get_object()
        self.perform_destoy(self.object)
        return HttpResponseRedirect(self.get_destroy_success_url())

    def perform_destoy(self, obj):
        """Performs the deletion operation."""
        session = self.get_session()
        session.delete(obj)
        session.flush()

    def get_destroy_context_data(self, **kwargs):
        """Returns destory context data for rendering deletion confirmation
        page."""
        return self.get_detail_context_data(**kwargs)

    def get_destroy_success_url(self):
        """Return the url to redirect to after successful deletion of an
        object."""
        if self.destroy_success_url:
            return self.destroy_success_url.format(**vars(self.object))

        raise ImproperlyConfigured(
            "No URL to redirect to. Either provide a url or override this function to return a url"
        )



"""InlineFormSet for sqlalchemy."""

from ..db import meta
from ..forms import ModelForm
from .base import BaseModelFormSet, modelformset_factory


class BaseInlineFormSet(BaseModelFormSet):
    """A formset for child objects related to a parent."""

    def __init__(self, data=None, files=None, instance=None, save_as_new=False, prefix=None, queryset=None, **kwargs):
        self.instance = self.fk.parent_model() if instance is None else instance
        self.save_as_new = save_as_new
        if queryset is None:
            queryset = getattr(self.instance, self.fk.name, [])

        super().__init__(data, files, prefix=prefix, queryset=queryset, **kwargs)

    @classmethod
    def get_default_prefix(cls):
        return cls.fk.name

    def initial_form_count(self):
        return 0 if self.save_as_new else super().initial_form_count()

    def save(self, flush=False, **kwargs):
        instances = super().save(flush=flush)
        setattr(self.instance, self.fk.name, instances)
        return instances


def _get_relation_info(relation):
    info = meta.model_info(relation.parent)
    rel = info.relationships[relation.key]
    if rel.uselist:
        return rel

    raise ValueError(f"Relationship '{relation}' is not one to many or many to many.")


def _get_foreign_key(relation, parent_model, model, fk_name=None):
    info = meta.model_info(parent_model)
    if fk_name is None:
        relations = [
            relation for relation in info.relationships.values() if relation.related_model == model and relation.uselist
        ]
        if len(relations) == 1:
            return relations[0]

        raise ValueError(f"Couldn't find a relation from '{parent_model.__name__}' to '{model.__name__}'.")


def inlineformset_factory(
    parent_model=None,
    model=None,
    relation=None,
    form=ModelForm,
    formset=BaseInlineFormSet,
    fk_name=None,
    fields=None,
    exclude=None,
    extra=3,
    can_order=False,
    can_delete=True,
    max_num=None,
    formfield_callback=None,
    widgets=None,
    validate_max=False,
    localized_fields=None,
    labels=None,
    help_texts=None,
    error_messages=None,
    min_num=None,
    validate_min=False,
    field_classes=None,
    session=None,
):
    """Return an ``InlineFormSet`` for the given kwargs.

    ``fk_name`` must be provided if ``model`` has more than one
    ``ForeignKey`` to ``parent_model``.
    """
    if relation is not None:
        fk = _get_relation_info(relation)
        parent_model = fk.parent_model
        model = fk.related_model
    else:
        fk = _get_foreign_key(relation, parent_model, model, fk_name=fk_name)

    kwargs = {
        "form": form,
        "formfield_callback": formfield_callback,
        "formset": formset,
        "extra": extra,
        "can_delete": can_delete,
        "can_order": can_order,
        "fields": fields,
        "exclude": exclude,
        "min_num": min_num,
        "max_num": max_num,
        "widgets": widgets,
        "validate_min": validate_min,
        "validate_max": validate_max,
        "localized_fields": localized_fields,
        "labels": labels,
        "help_texts": help_texts,
        "error_messages": error_messages,
        "field_classes": field_classes,
        "session": session,
    }
    FormSet = modelformset_factory(model, **kwargs)
    FormSet.fk = fk
    return FormSet



import enum
from datetime import date, datetime, time, timedelta
from decimal import Decimal

import pytz
import sqlalchemy as sa
from django import forms as djangoforms
from django.core import validators as django_validators
from django.core.exceptions import ValidationError
from django.forms import fields as djangofields, widgets
from django_sorcery import fields as sorceryfields
from django_sorcery.db import fields as dbfields, meta

from ...base import TestCase, mock
from ...testapp.models import (
    COLORS,
    AllKindsOfFields,
    Business,
    Vehicle,
    VehicleType,
)


class TestColumnMeta(TestCase):
    def test_column_info(self):
        info = meta.model_info(Vehicle)

        col = info.primary_keys["id"]
        self.assertEqual(col.name, "id")
        self.assertFalse(col.null)
        self.assertDictEqual(
            {
                "help_text": "The primary key",
                "label": "Id",
                "required": True,
                "validators": [django_validators.validate_integer, mock.ANY, mock.ANY],
            },
            col.field_kwargs,
        )
        self.assertIs(col.validators, col.column.info["validators"])
        self.assertTrue(col.required)
        self.assertIs(Vehicle.id, col.attribute)

    def test_default_initial(self):
        info = meta.model_info(Business)

        col = info.employees
        self.assertDictEqual(
            {
                "help_text": None,
                "initial": 5,
                "label": "Employees",
                "required": True,
                "validators": [django_validators.validate_integer, mock.ANY, mock.ANY],
            },
            col.field_kwargs,
        )

    def test_column_info_enum(self):
        info = meta.model_info(Vehicle)
        col = info.type
        self.assertFalse(col.null)
        self.assertDictEqual(
            {"choices": VehicleType, "help_text": None, "label": "Type", "required": True, "validators": []},
            col.field_kwargs,
        )
        self.assertTrue(col.required)

    def test_column_info_enum_from_class_from_info(self):
        col = sa.Column(sa.Enum(VehicleType), info={"form_class": djangofields.IntegerField})
        info = meta.column_info(col, name="test")
        self.assertIsInstance(info, meta.enum_column_info)
        self.assertEqual(info.form_class, djangofields.IntegerField)

    def test_column_info_boolean_from_class_from_info(self):
        info = meta.column_info(sa.Column(sa.Boolean(), info={"form_class": djangofields.IntegerField}), name="test")
        self.assertIsInstance(info, meta.boolean_column_info)
        self.assertEqual(info.form_class, djangofields.IntegerField)

    def test_column_info_validators(self):
        info = meta.model_info(Vehicle)
        col = info.properties["msrp"]
        validator = col.column.info["validators"][0]
        self.assertIsInstance(validator, django_validators.DecimalValidator)
        self.assertDictEqual(
            {
                "decimal_places": 2,
                "help_text": None,
                "label": "Msrp",
                "max_digits": 10,
                "required": False,
                "validators": [validator],
            },
            col.field_kwargs,
        )

    def test_column_info_label_and_text(self):
        column = meta.column_info(dbfields.TextField(label="dummy"), name="test")
        self.assertDictEqual(
            column.field_kwargs,
            {
                "help_text": None,
                "label": "dummy",
                "max_length": None,
                "required": False,
                "validators": [],
                "widget": djangoforms.Textarea,
            },
        )

    def test_choice_enum(self):
        info = meta.model_info(Vehicle)
        col = info.paint
        self.assertDictEqual(
            {
                "choices": [(x, x) for x in COLORS],
                "help_text": None,
                "label": "Paint",
                "required": False,
                "validators": [],
            },
            col.field_kwargs,
        )
        self.assertEqual(col.parent_model, Vehicle)
        self.assertEqual(repr(col), "<choice_column_info(Vehicle.paint)>")

    def test_plain_sqla(self):
        info = meta.model_info(AllKindsOfFields)
        col = info.decimal
        self.assertDictEqual(
            {"help_text": None, "label": "Decimal", "required": False, "validators": []}, col.field_kwargs
        )


def _run_tests(test, column_info, tests):
    for value, exp in tests:
        instance = object()
        if exp is ValidationError:
            with test.assertRaises(ValidationError):
                column_info.clean(value, instance)
        else:
            test.assertEqual(column_info.clean(value, instance), exp)


class TestDefaultColumn(TestCase):
    def test_formfield(self):
        info = meta.column_info(sa.Column(sa.sql.sqltypes.ARRAY(item_type=int)), name="test")
        formfield = info.formfield()
        self.assertIsNone(formfield)

    def test_clean(self):
        info = meta.column_info(sa.Column(sa.sql.sqltypes.ARRAY(item_type=int)), name="test")
        tests = [("test", "test"), (1, 1), (None, None)]
        _run_tests(self, info, tests)


class TestStringColumn(TestCase):
    def test_formfield(self):
        info = meta.column_info(sa.Column(sa.String()), name="test")
        formfield = info.formfield()
        self.assertIsInstance(formfield, djangofields.CharField)

        info = meta.column_info(sa.Column(sa.Text()), name="test")
        formfield = info.formfield()
        self.assertIsInstance(formfield, djangofields.CharField)
        self.assertIsInstance(formfield.widget, widgets.Textarea)

    def test_clean(self):
        info = meta.column_info(sa.Column(sa.String()), name="test")

        tests = [
            (None, None),
            ("abc", "abc"),
            (1234, "1234"),
            ("", ""),
            ("é", "é"),
            (Decimal("1234.44"), "1234.44"),
            # excess whitespace tests
            ("\t\t\t\t\n", ""),
            ("\t\tabc\t\t\n", "abc"),
            ("\t\t20,000\t\t\n", "20,000"),
            ("  \t 23\t", "23"),
        ]
        _run_tests(self, info, tests)


class TestChoiceColumn(TestCase):
    def test_formfield(self):
        info = meta.column_info(sa.Column(sa.Enum(*["1", "2", "3"])), name="test")
        formfield = info.formfield()
        self.assertIsInstance(formfield, djangofields.TypedChoiceField)

    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Enum(*["1", "2", "3"])), name="test")
        tests = [(None, None), ("", None), ("1", "1"), (1, "1"), ("4", ValidationError), (5, ValidationError)]
        _run_tests(self, info, tests)


class TestEnumColumn(TestCase):
    def test_formfield(self):
        class Demo(enum.Enum):
            one = "1"
            two = "2"

        info = meta.column_info(sa.Column(sa.Enum(Demo)), name="test")
        formfield = info.formfield()
        self.assertIsInstance(formfield, sorceryfields.EnumField)

    def test_clean(self):
        class Demo(enum.Enum):
            one = "1"
            two = "2"

        info = meta.column_info(sa.Column(sa.Enum(Demo)), name="test")
        tests = [
            (None, None),
            ("", None),
            ("one", Demo.one),
            ("1", Demo.one),
            (Demo.one, Demo.one),
            (1, ValidationError),
        ]
        _run_tests(self, info, tests)


class TestNumericColumn(TestCase):
    def test_formfield(self):
        info = meta.column_info(sa.Column(sa.Numeric(precision=14, scale=2, asdecimal=True)), name="test")
        formfield = info.formfield()
        self.assertIsInstance(formfield, djangofields.DecimalField)
        self.assertEqual(formfield.max_digits, 14)
        self.assertEqual(formfield.decimal_places, 2)

    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Numeric(precision=14, scale=2, asdecimal=True)), name="test")
        tests = [
            (None, None),
            ("", None),
            ("1", Decimal("1")),
            (Decimal("1"), Decimal("1")),
            ("abc", ValidationError),
            (1, Decimal("1")),
            (4123411130.3419398, Decimal("4123411130.3419")),
            ("20,000", Decimal("20000")),
            ("1.e-8", Decimal("1E-8")),
            ("1.-8", ValidationError),
            # excess whitespace tests
            ("\t\t\t\t\n", None),
            ("\t\tabc\t\t\n", ValidationError),
            ("\t\t20,000\t\t\n", Decimal("20000")),
            ("  \t 23\t", Decimal("23")),
        ]
        _run_tests(self, info, tests)

        with self.settings(THOUSAND_SEPARATOR=".", DECIMAL_SEPARATOR=","):
            tests = [("20.000", Decimal("20000")), ("20,000", Decimal("20.000"))]
            _run_tests(self, info, tests)


class TestBooleanColumn(TestCase):
    def test_formfield(self):
        info = meta.column_info(sa.Column(sa.Boolean()), name="test")
        formfield = info.formfield()
        self.assertIsInstance(formfield, djangofields.NullBooleanField)

        info = meta.column_info(sa.Column(sa.Boolean(), nullable=False), name="test")
        formfield = info.formfield()
        self.assertIsInstance(formfield, djangofields.BooleanField)

    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Boolean()), name="test")
        tests = [
            (None, None),
            ("", None),
            (True, True),
            (1, True),
            ("t", True),
            ("true", True),
            ("True", True),
            ("1", True),
            (0, False),
            ("f", False),
            ("false", False),
            ("False", False),
            ("0", False),
        ]
        _run_tests(self, info, tests)


class TestDateColumn(TestCase):
    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Date()), name="test")
        tests = [
            (None, None),
            ("Hello", ValidationError),
            (date(2006, 10, 25), date(2006, 10, 25)),
            (datetime(2006, 10, 25, 14, 30), date(2006, 10, 25)),
            (datetime(2006, 10, 25, 14, 30, 59), date(2006, 10, 25)),
            (datetime(2006, 10, 25, 14, 30, 59, 200), date(2006, 10, 25)),
            (datetime(2006, 10, 25, 14, 30, 59, 200, tzinfo=pytz.timezone("EST")), date(2006, 10, 25)),
            ("2006-10-25", date(2006, 10, 25)),
            ("10/25/2006", date(2006, 10, 25)),
            ("10/25/06", date(2006, 10, 25)),
            ("Oct 25 2006", date(2006, 10, 25)),
            ("October 25 2006", date(2006, 10, 25)),
            ("October 25, 2006", date(2006, 10, 25)),
            ("25 October 2006", date(2006, 10, 25)),
            ("25 October, 2006", date(2006, 10, 25)),
            (1335172500.12, date(2012, 4, 23)),
        ]
        _run_tests(self, info, tests)


class TestDateTimeColumn(TestCase):
    def test_clean(self):
        info = meta.column_info(sa.Column(sa.DateTime()), name="test")
        tests = [
            (None, None),
            (
                datetime(2006, 10, 25, 14, 30, 45, 200, tzinfo=pytz.timezone("EST")),
                datetime(2006, 10, 25, 19, 30, 45, 200),
            ),
            (date(2006, 10, 25), datetime(2006, 10, 25, 0, 0)),
            ("2006-10-25 14:30:45.000200", datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("2006-10-25 14:30:45.0002", datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("2006-10-25 14:30:45", datetime(2006, 10, 25, 14, 30, 45)),
            ("2006-10-25 14:30:00", datetime(2006, 10, 25, 14, 30)),
            ("2006-10-25 14:30", datetime(2006, 10, 25, 14, 30)),
            ("2006-10-25", datetime(2006, 10, 25, 0, 0)),
            ("10/25/2006 14:30:45.000200", datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("10/25/2006 14:30:45", datetime(2006, 10, 25, 14, 30, 45)),
            ("10/25/2006 14:30:00", datetime(2006, 10, 25, 14, 30)),
            ("10/25/2006 14:30", datetime(2006, 10, 25, 14, 30)),
            ("10/25/2006", datetime(2006, 10, 25, 0, 0)),
            ("10/25/06 14:30:45.000200", datetime(2006, 10, 25, 14, 30, 45, 200)),
            ("10/25/06 14:30:45", datetime(2006, 10, 25, 14, 30, 45)),
            ("10/25/06 14:30:00", datetime(2006, 10, 25, 14, 30)),
            ("10/25/06 14:30", datetime(2006, 10, 25, 14, 30)),
            ("10/25/06", datetime(2006, 10, 25, 0, 0)),
            ("2012-04-23T09:15:00", datetime(2012, 4, 23, 9, 15)),
            ("2012-4-9 4:8:16", datetime(2012, 4, 9, 4, 8, 16)),
            ("2012-04-23T09:15:00Z", datetime(2012, 4, 23, 9, 15, 0, 0)),
            ("2012-4-9 4:8:16-0320", datetime(2012, 4, 9, 7, 28, 16, 0)),
            ("2012-04-23T10:20:30.400+02:30", datetime(2012, 4, 23, 7, 50, 30, 400000)),
            ("2012-04-23T10:20:30.400+02", datetime(2012, 4, 23, 8, 20, 30, 400000)),
            ("2012-04-23T10:20:30.400-02", datetime(2012, 4, 23, 12, 20, 30, 400000)),
            (1335172500.12, datetime(2012, 4, 23, 9, 15, 0, 120000)),
            ("Hello", ValidationError),
        ]
        _run_tests(self, info, tests)


class TestFloatColumn(TestCase):
    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Float()), name="test")
        tests = [
            (None, None),
            ("", None),
            (1.2, 1.2),
            ("1", 1.0),
            ("abc", ValidationError),
            ("1.0", 1.0),
            ("1.", 1.0),
            ("1.001", 1.001),
            ("1.e-8", 1e-08),
            ("\t\t\t\t\n", None),
            ("\t\tabc\t\t\n", ValidationError),
            ("\t\t20,000.02\t\t\n", 20000.02),
            ("  \t 23\t", 23.0),
        ]
        _run_tests(self, info, tests)


class TestIntegerColumn(TestCase):
    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Integer()), name="test")
        tests = [
            (None, None),
            ("", None),
            (1, 1),
            ("1", 1),
            ("abc", ValidationError),
            (1.0, 1),
            (1.01, ValidationError),
            ("1.0", 1),
            ("1.01", ValidationError),
            ("\t\t\t\t\n", None),
            ("\t\tabc\t\t\n", ValidationError),
            ("\t\t20,000.02\t\t\n", ValidationError),
            ("\t\t20,000\t\t\n", 20000),
            ("  \t 23\t", 23),
        ]
        _run_tests(self, info, tests)


class TestIntervalColumn(TestCase):
    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Interval()), name="test")
        tests = [
            (None, None),
            ("", None),
            ("Hello", ValidationError),
            (1.2, timedelta(seconds=1, microseconds=200000)),
            (timedelta(seconds=30), timedelta(seconds=30)),
            ("30", timedelta(seconds=30)),
            ("15:30", timedelta(minutes=15, seconds=30)),
            ("1:15:30", timedelta(hours=1, minutes=15, seconds=30)),
            ("1 1:15:30.3", timedelta(days=1, hours=1, minutes=15, seconds=30, milliseconds=300)),
        ]
        _run_tests(self, info, tests)


class TestTimeColumn(TestCase):
    def test_clean(self):
        info = meta.column_info(sa.Column(sa.Time()), name="test")
        tests = [
            (None, None),
            ("", None),
            ("Hello", ValidationError),
            (2.3, ValidationError),
            (time(14, 25), time(14, 25)),
            (time(14, 25, 59), time(14, 25, 59)),
            (datetime(2006, 10, 25, 14, 30, 59, 200, tzinfo=pytz.timezone("EST")), time(14, 30, 59, 200)),
            ("14:25", time(14, 25)),
            ("14:25:59", time(14, 25, 59)),
        ]
        _run_tests(self, info, tests)



"""Default alembic signal configuration."""

from ...db import signals


def include_object(obj, name, type_, reflected, compare_to):
    """The default include_object handler for alembic, delegates to
    alembic_include_object signal.

    The return value will be considered True if all the signal handlers
    return True for the object to be included in migration.
    """
    results = signals.alembic_include_object.send(
        obj, name=name, type_=type_, reflected=reflected, compare_to=compare_to
    )
    if results:
        return all(res[1] for res in results)

    return True


def process_revision_directives(context, revision, directives):
    """The default process_revision_directives handler for alembic, delegates
    to alembic_process_revision_directives."""
    signals.alembic_process_revision_directives.send(context, revision=revision, directives=directives)



"""Validators."""

import sqlalchemy as sa
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from sqlalchemy import inspect


class ValidateTogetherModelFields:
    """Validator for checking that multiple model fields are always saved
    together.

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.Integer())
            bar = db.Column(db.Integer())

            validators = [
                ValidateTogetherModelFields(["foo", "bar"]),
            ]
    """

    message = _("All %(fields)s are required.")
    code = "required"

    def __init__(self, fields, message=None, code=None):
        self.fields = fields
        self.message = message or self.message
        self.code = code or self.code

    def __call__(self, m):
        if not any(getattr(m, i, None) for i in self.fields):
            return
        if not all(getattr(m, i, None) for i in self.fields):
            raise ValidationError(self.message, code=self.code, params={"fields": ", ".join(sorted(self.fields))})


class ValidateUnique:
    """Validator for checking uniqueness of arbitrary list of attributes on a
    model.

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.Integer())
            bar = db.Column(db.Integer())
            name = db.Column(db.Integer())

            validators = [
                ValidateUnique(db, "name"),      # checks for name uniqueness
                ValidateUnique(db, "foo", "bar"),  # checks for foo and bar combination uniqueness
            ]
    """

    message = _("%(fields)s must make a unique set.")
    code = "required"

    def __init__(self, session, *args, **kwargs):
        self.session = session
        self.message = kwargs.get("message", self.message)
        self.code = kwargs.get("code", self.code)
        self.attrs = args

    def __call__(self, m):
        clauses = [getattr(m.__class__, attr) == getattr(m, attr) for attr in self.attrs]

        from ..db import meta

        info = meta.model_info(m)
        state = info.sa_state(m)
        if state.persistent:
            # need to exlude the current model since it's already in db
            pks = info.mapper.primary_key_from_instance(m)
            for name, pk in zip(info.primary_keys, pks):
                clauses.append(getattr(m.__class__, name) != pk)

        query = self.session.query(m.__class__).filter(*clauses)
        exists = self.session.query(sa.literal(True)).filter(query.exists()).scalar()

        if exists:
            raise ValidationError(self.message, code=self.code, params={"fields": ", ".join(sorted(self.attrs))})


class ValidateValue:
    """Validator for checking correctness of a value by using a predicate
    callable.

    Useful when other multiple fields need to be consulted
    to check if particular field is valid.

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.String())
            bar = db.Column(db.String())

            validators = [
                # allow only "bar" value when model.foo == "foo"
                ValidateValue('bar', lambda m: m.foo != 'foo' or b.bar == 'bar'),
            ]
    """

    message = _("Please enter valid value.")
    code = "invalid"

    negated = False

    def __init__(self, field, predicate, message=None, code=None):
        assert callable(predicate), "predicate must be callable"
        self.field = field
        self.predicate = predicate
        self.message = message or self.message
        self.code = code or self.code

    def __call__(self, m):
        e = ValidationError({self.field: ValidationError(self.message, code=self.code, params={"field": self.field})})
        try:
            is_valid = self.predicate(m)
        except Exception:
            raise e
        else:
            if not is_valid:
                raise e


class ValidateEmptyWhen:
    """Validator for checking a field is empty when predicate is True.

    Useful to conditionally enforce a field is not provided depending on related field

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.String())
            bar = db.Column(db.String())

            validators = [
                # do not allow to set bar unless foo is present
                ValidateEmptyWhen('bar', lambda m: not m.foo),
            ]
    """

    message = _("Cannot provide a value to this field.")
    code = "empty"

    allow_empty = True

    def __init__(self, field, predicate, message=None, code=None):
        assert callable(predicate), "predicate must be callable"
        self.field = field
        self.predicate = predicate
        self.message = message or self.message
        self.code = code or self.code

    def __call__(self, m):
        is_empty = not bool(getattr(m, self.field, None))

        if (self.allow_empty and is_empty) or (not self.allow_empty and not is_empty):
            return

        if self.predicate(m):
            raise ValidationError(
                {self.field: ValidationError(self.message, code=self.code, params={"field": self.field})}
            )


class ValidateNotEmptyWhen(ValidateEmptyWhen):
    """Validator for checking a field is provided when predicate is True.

    Useful to conditionally enforce a field is provided depending on related field

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.String())
            bar = db.Column(db.String())

            validators = [
                # enforce bar is provided when foo is provided
                ValidateNotEmptyWhen('bar', lambda m: m.foo),
            ]
    """

    message = _("This field is required.")
    code = "not_empty"

    allow_empty = False


class ValidateOnlyOneOf:
    """Validate that only one of given fields is provided.

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.String())
            bar = db.Column(db.String())

            validators = [
                # enforce only either foo or bar are provided
                ValidateOnlyOneOf(['foo', 'bar']),
            ]
    """

    message = _("Only one of %(fields)s is allowed.")
    code = "exclusive"

    def __init__(self, fields, required=True, message=None, code=None):
        self.fields = fields
        self.required = required
        self.message = message or self.message
        self.code = code or self.code

    def __call__(self, m):
        e = ValidationError(self.message, code=self.code, params={"fields": ", ".join(sorted(self.fields))})

        present = len(list(filter(None, (getattr(m, i, None) for i in self.fields))))

        if (self.required and not present) or present > 1:
            raise e


class ValidateCantRemove:
    """Validate that for data cannot be removed for given field.

    For example::

        class MyModel(db.Model):
            foo = db.Column(db.String())

            validators = [
                ValidateCantRemove('foo'),
            ]
    """

    message = _("Cannot remove existing data.")
    code = "remove"

    def __init__(self, field, message=None, code=None):
        self.field = field
        self.message = message or self.message
        self.code = code or self.code

    def __call__(self, m):
        inspected = inspect(m).attrs
        history = getattr(inspected, self.field).history

        if not history.has_changes():
            return

        previous, current = next(iter(history.deleted or ()), None), getattr(m, self.field)
        if previous is not None and current is None:
            raise ValidationError(
                {self.field: ValidationError(self.message, code=self.code, params={"field": self.field})}
            )
